AllaraBossCore = LibStub("AceAddon-3.0"):NewAddon("AllaraBossCore", "AceConsole-3.0", "AceEvent-3.0", "LibShefkiTimer-1.0", "AceComm-3.0", "AceSerializer-3.0")
AllaraBossCore.MINOR_VERSION = tonumber(("$Revision: 29 $"):match("%d+"))
local L = LibStub("AceLocale-3.0"):GetLocale("AllaraBossCore")
local ABC = AllaraBossCore
local zonenames = L
--


--[[
      VARIABLES ----------------------------------------------------------------------------------------------------------------
  ]]

local POLLING_FREQUENCY = 1.0
local targetList = {}
local pullChecks = 0
local commPrefix = "AllaraBossCore"
local startingFight = false

ABC.isEnabled = false
ABC.inCombat = false
ABC.ver = {}
ABC.raid = {}
ABC.enc = {}
ABC.zones = {}
ABC.fight = {}
ABC.modules = {}
ABC.testZone = nil
ABC.forceHeroic = false
ABC.testing = false
ABC.forcedTargetNpc = nil
ABC.forcedTargetName = nil

StaticPopupDialogs["ALLARABOSSCORE_UPGRADE_NOTICE"] = {
	text = "",
	button1 = ACCEPT,
	hideOnEscape = 1,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
}



--[[
      UPVALUES -----------------------------------------------------------------------------------------------------------------
  ]]
  
local _G = getfenv(0)
local pairs = _G.pairs
local strlower = _G.strlower
local strupper = _G.strupper
local format = _G.format

local UnitAffectingCombat = _G.UnitAffectingCombat
local UnitGUID = _G.UnitGUID
local UnitName = _G.UnitName
local UnitInRaid = _G.UnitInRaid
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost



--[[
      INITIALIZATION -----------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnInitialize()
	self:RegisterChatCommand("allarabosscore", "ChatCommand")
	if not AllaraBossCoreDB then AllaraBossCoreDB = {} end
end


function ABC:OnEnable()
	-- Scan for add-ons
	self.modules = {}
	for i = 1, GetNumAddOns() do
		if GetAddOnMetadata(i, "X-AllaraBossCore-LoadZone") then
			table.insert(self.modules, {
				addon = GetAddOnInfo(i),
				zone = {strsplit(",", GetAddOnMetadata(i, "X-AllaraBossCore-LoadZone") or "")},
				minrev = GetAddOnMetadata(i, "X-AllaraBossCore-MinimumRevision") or 1
			})
		end
	end
	local noticed = false
	for k, v in pairs(self.modules) do
		if v.minrev and (tonumber(v.minrev) or 1) > ABC.MINOR_VERSION then
			self:Print(format(L["|cffffff20%s|r requires Allara Boss Core revision |cffffff20%d|r or higher. It will not function until you update."], v.addon, v.minrev))
			if not noticed then
				StaticPopupDialogs["ALLARABOSSCORE_UPGRADE_NOTICE"].text = format(L["|cffffff20%s|r requires Allara Boss Core revision |cffffff20%d|r or higher. It will not function until you update."].."\n\n"..L["Please update by navigating to:"].."\n\n%s",
					v.addon, v.minrev, "http://www.wowace.com/addons/allara-boss-core/")
				StaticPopup_Show("ALLARABOSSCORE_UPGRADE_NOTICE")
				noticed = true
			end
			self.modules[k] = nil
		end
	end

	-- Register for events
	self:UnregisterAllEvents()
	self:RegisterEvent("RAID_ROSTER_UPDATE", "OnEvent")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnEvent")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnEvent")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_YELL", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE", "OnEvent")
	self:RegisterEvent("CHAT_MSG_RAID_BOSS_EMOTE", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_SAY", "OnEvent")

	-- Register communication
	self:UnregisterAllComm()
 self:RegisterComm(commPrefix, "OnCommReceived")

	-- Initialize
	self:LoadZone()
	self:UpdateRaid()
	
	-- Turn us on
	ABC.isEnabled = true
	
	-- Ask the raid if we should be in a fight right now
	if UnitInRaid("player") then
		self:SendCommMessage(commPrefix, self:Serialize({ c = "FIGHT_REQUEST" }), "RAID")
	end

	self:Debug("Allara Boss Core loaded (running in debug mode)")
end


function ABC:OnDisable()
	self:EndAllFights()

	-- Cancel all timers
	self:CancelAllTimers()

	-- Turn us off
	ABC.isEnabled = false
	
	self:Debug(L["Disabled"])
end


function ABC:ChatCommand(input)
	if strlower(input) == "ver" then
		self:VersionCheck()
	elseif strlower(input) == "debug" then
		if AllaraBossCoreDB.debugMode then
			self:Print("Debug mode is now OFF")
			AllaraBossCoreDB.debugMode = false
		else
			self:Print("Debug mode is now ON")
			AllaraBossCoreDB.debugMode = true
		end
	end
end


function ABC:VersionCheck()
	if UnitInRaid("player") then
		ver = {}
		self:SendCommMessage(commPrefix, self:Serialize({ c = "VER_REQ" }), "RAID")
		self:ScheduleTimer("OnVersionCheckFinished", 5)
	end
end

function ABC:OnVersionCheckFinished()
	if UnitInRaid("player") then
		for k, v in pairs(ABC.raid) do
			if not ver[k] and UnitInRaid(k) and v.online then
				self:Print(format(L["%s is not using Allara Boss Core"], k))
			end
		end
		local count = 0
		for k, v in pairs(ver) do count = count + 1 end
		local numNot = GetNumRaidMembers() - count
		if numNot < 0 then numNot = 0 end
		self:Print(format(L["%d player(s) are running Allara Boss Core. %d player(s) are not."], count, numNot))
	end
end



--[[
      EVENTS -------------------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnEvent(event, ...)

	if event == "RAID_ROSTER_UPDATE" then
		self:UpdateRaid()
		
	elseif event == "ZONE_CHANGED_NEW_AREA" then
		self:LoadZone()
						
	elseif event == "PLAYER_REGEN_DISABLED" then
		if UnitInRaid("player") then
			self.inCombat = true
			self:CheckForPull()
		end
			
	elseif event == "PLAYER_REGEN_ENABLED" then
		self.inCombat = false
		self:UpdateRaid()
		
	elseif event == "CHAT_MSG_MONSTER_YELL" then
		local msg = select(1, ...)
		for k, v in pairs(ABC.fight) do
			if v.onYell and type(v.onYell) == "function" then v.onYell(v, event, ...) end
		end
		self:OnMonsterMessage("yell", msg)
	
	elseif event == "CHAT_MSG_MONSTER_EMOTE" then
		local msg = select(1, ...)
		for k, v in pairs(ABC.fight) do
			if v.onYell and type(v.onYell) == "function" then v.onYell(v, event, ...) end
		end
		self:OnMonsterMessage("emote", msg)
	
	elseif event == "CHAT_MSG_RAID_BOSS_EMOTE" then
		local msg = select(1, ...)
		for k, v in pairs(ABC.fight) do
			if v.onYell and type(v.onYell) == "function" then v.onYell(v, event, ...) end
		end
		self:OnMonsterMessage("emote", msg)
	
	elseif event == "CHAT_MSG_MONSTER_SAY" then
		local msg = select(1, ...)
		for k, v in pairs(ABC.fight) do
			if v.onYell and type(v.onYell) == "function" then v.onYell(v, event, ...) end
		end
		self:OnMonsterMessage("say", msg)
					
 end
end


function ABC:OnCombatEvent(event, timestamp, eventType, hideCaster, srcGuid, srcName, srcFlags, srcRaidFlags, dstGuid, dstName, dstFlags, dstRaidFlags, ...)
	if ABC.isEnabled then
		
		for k, v in pairs(ABC.fight) do
			if v.combatEvents and (v.combatEvents.ALL or v.combatEvents[eventType]) then
				if v.onCombat and type(v.onCombat) == "function" then
					v.onCombat(v, event, timestamp, eventType, hideCaster, srcGuid, srcName, srcFlags, srcRaidFlags, dstGuid, dstName, dstFlags, dstRaidFlags, ...)
				end
			end
		end
		
		if eventType == "UNIT_DIED" or eventType == "UNIT_DESTROYED" then
			if bit.band(dstGuid:sub(1, 5), 0x00F) == 3 or bit.band(dstGuid:sub(1, 5), 0x00F) == 5 then
				self:OnMobKill(tonumber(dstGuid:sub(9, 12), 16))
			end
		end

	end
end


function ABC:OnCommReceived(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived] Failed to deserialize message")
  return
 else
		if o.c then
		
			if o.c == "VER" then
				if o.v then
					ver[sender] = o.v
					self:Print(format(L["%s is using Allara Boss Core r%s"], sender, o.v))
				end

			elseif o.c == "VER_REQ" then
				self:SendCommMessage(commPrefix, self:Serialize({ c = "VER", v = self.MINOR_VERSION }), "WHISPER", sender)

			elseif o.c == "FIGHT_REQUEST" and sender ~= UnitName("player") then
				for k, v in pairs(ABC.fight) do
					self:SendCommMessage(commPrefix, self:Serialize({ c = "FIGHT_RESPONSE", f = v._key }), "WHISPER", sender)
				end
				
			elseif o.c == "FIGHT_RESPONSE" and sender ~= UnitName("player") then
					if ABC.enc[o.f] then
						if not ABC:IsFightStarted(ABC.enc[o.f]) then
							self:LoadZone(nil, true)
							self:ScheduleTimer(function()
								self:StartFight(ABC.enc[o.f], 0)
								startingFight = false
							end, 3.0)
						end
					end
				
			elseif o.c == "SYNC" then
				if o.addon and o.data and sender ~= UnitName("player") then
					for k, v in pairs(ABC.fight) do
						if v.onSync and type(v.onSync) == "function" and v._key == o.addon then
							v.onSync(v, o.data)
						end
					end
				end
								
			end
			
		end
 end
end



--[[
      TARGET DETECTION ---------------------------------------------------------------------------------------------------------
  ]]
  
do

	--[[
	
		GetBossTargetChange (typically useful for SPELL_CAST_START events during
		really picky mechanics):
		
			If the boss is your target or focus, start watching for UNIT_TARGET
			events. If one is found, use the unit's target at that point. If one
			is not found within 0.3 seconds, use the target at 0.3 seconds.
			
			If the boss isn't your target or focus, scan the raid to find it.
			Save it's target's GUID. Scan every 0.05 seconds for 0.3 seconds and see if
			the boss changes targets during that time. If so, use the new target.
			If not, use the target at 0.3 seconds.
			
		GetBossTarget (typically useful for SPELL_CAST_COMPLETE events):
		
		 After a configurable delay, scans your target, focus, and the entire
		 raid to find the boss. Once found, uses the boss target.
			
	]]

	local isScanning = false
	local lastGuid
	local interested_unit
	local tries = 0
	local tryTimer, cancelTimer
	local FAILSAFE_TIME = 0.3
	local TRY_REPEAT_TIME = 0.05
	local MAX_TRIES = 6
	local fn, cid
	local inDoTarget = false
	
	
	local function deinit()
		if tryTimer then ABC:CancelTimer(tryTimer, true) end
		if cancelTimer then ABC:CancelTimer(cancelTimer, true) end
		isScanning = false
		lastGuid = nil
		interested_unit = nil
		fn = nil
		cid = nil
		tryTimer = nil
		cancelTimer = nil
	end
	
	local function doTarget(unit)
		if not inDoTarget then
			inDoTarget = true
			local u = unit.."target"
			if UnitExists(u) then
				if fn and type(fn) == "function" then
					local n = UnitName(u)
					if n then
						fn(n)
					end
				end
			end
			deinit()
			inDoTarget = false
		end
	end


	function ABC:OnTarget(event, unit) -- Fires for UNIT_TARGET events
		if isScanning and unit == interested_unit then
			doTarget(unit)
		end
	end
	
	
	-- failsafe doesn't get scheduled for a normal repeated scan
	local function failsafe()
		if interested_unit then
			doTarget(interested_unit)
		else
			local name, unit, guid = scan(cid)
			if unit then
				doTarget(unit)
			else
				deinit()
			end
		end
	end
	
	
	local t = {"target", "targettarget", "focus", "focustarget", "mouseover", "mouseovertarget"}
	for i = 1, 4 do t[#t+1] = format("boss%d", i) end
	for i = 1, 4 do t[#t+1] = format("party%dtarget", i) end
	for i = 1, 40 do t[#t+1] = format("raid%dtarget", i) end
	local function findBoss(id)
		local idType = type(id)
		for i, unit in next, t do
			if UnitExists(unit) and not UnitIsPlayer(unit) then
				local unitId = UnitGUID(unit)
				if idType == "number" then unitId = tonumber(unitId:sub(-12, -7), 16) end
				if unitId == id then return unit end
			end
		end
	end
	

	local function scan(id)
		local unit = findBoss(id)
		if not unit then return end
		if UnitExists(unit.."target") then return UnitName(unit.."target"), unit, UnitGUID(unit.."target") end
	end
	
	
	local function try()
		if tries >= MAX_TRIES then return end
		tries = tries + 1
		local name, unit, guid = scan(cid)
		if unit and guid and guid ~= lastGuid then
			doTarget(unit)
			return
		end
		
		if tries == MAX_TRIES then
			if unit then
				doTarget(unit)
			else
				deinit()
			end
		end
	end


	--- Watches a boss to see if it changes targets within 0.3 seconds.
	-- If it does, your callback function is called. If the boss doesn't
	-- change targets, whatever it's target is after 0.3 seconds is used.
	-- This algorithm is necessary for certain boss mechanics.
	-- @param _cid The NPC ID of the boss.
	-- @param _fn The callback function to execute. The only parameter to this function will be the name of the player that is targetted.
	function ABC:GetBossTargetChange(_cid, _fn)
		deinit()
		inDoTarget = false
		
		fn = _fn
		cid = _cid

		if ABC.forcedTargetNpc == cid and ABC.forcedTargetName and fn then
			fn(ABC.forcedTargetName)
			return
		end
		
		if self:GetUnitCreatureId("target") == cid then
			interested_unit = "target"
			isScanning = true
			cancelTimer = self:ScheduleTimer(failsafe, FAILSAFE_TIME)
		elseif self:GetUnitCreatureId("focus") == cid then
			interested_unit = "focus"
			isScanning = true
			cancelTimer = self:ScheduleTimer(failsafe, FAILSAFE_TIME)
		else
			interested_unit = nil
			local name, unit, guid = scan(cid)
			lastGuid = guid
			tries = 0
			self:ScheduleRepeatingTimer(try, TRY_REPEAT_TIME)
		end
	end
	
	--- After a specified delay, locates the requested boss and finds it's target.
	-- @param _cid The NPC ID of the boss.
	-- @param _fn The callback function to execute. The only parameter to this function will be the name of the player that is targetted.
	-- @param _delay The delay in seconds before the target should be scanned for.
	function ABC:GetBossTarget(_cid, _fn, _delay)
		if ABC.forcedTargetNpc == _cid and ABC.forcedTargetName and _fn then
			_fn(ABC.forcedTargetName)
			return
		end

		self:ScheduleTimer(function()
			local name, unit, guid = scan(_cid)
			if name and type(_fn) == "function" then
				_fn(name)
			end
		end, _delay or 0.1)
	end
	
end



--[[
      PULL DETECTION -----------------------------------------------------------------------------------------------------------
  ]]
  
-- Combat-based detection (some raid member is targeting the boss, and the boss is affecting combat)
function ABC:BuildTargetList()
	local uId = "raid"
	for i = 0, GetNumRaidMembers() do
		local id = (i == 0 and "target") or uId..i.."target"
		local guid = UnitGUID(id)
		if guid then
			local cId = tonumber(guid:sub(7, 10), 16)
			targetList[cId] = id
		end
	end
end

function ABC:ClearTargetList()
	table.wipe(targetList)
end

function ABC:DoDelayedCheckForPull(mob, fight)
	if not fight.inCombat then
		ABC:BuildTargetList()
		pullChecks = pullChecks + 1
		if targetList[mob] and UnitAffectingCombat(targetList[mob]) then
			self:Debug("Pull detected")
			ABC:StartFight(fight, 3 * pullChecks)
		elseif targetList[mob] and pullChecks < 10 then
			ABC:ScheduleTimer(function() ABC:DoDelayedCheckForPull(mob, fight) end, 3)
		end
		ABC:ClearTargetList()
	end
end

function ABC:DoCheckForPull(mob, fight)
	local uId = targetList[mob]
	if uId and UnitAffectingCombat(uId) then
		self:Debug("Pull detected")
		ABC:StartFight(fight, 0)
		return true
	elseif uId then
		ABC:ScheduleTimer(function() ABC:DoDelayedCheckForPull(mob, fight) end, 3)
	end
end

function ABC:CheckForPull()
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		pullChecks = 0
		ABC:BuildTargetList()
		for k, v in pairs(ABC.enc) do
			if v.pull and v.pull.type == "combat" and v.pull.npcs then
				for kk, vv in pairs(v.pull.npcs) do
					if ABC:DoCheckForPull(vv, v) then break end
				end
			end
		end
		ABC:ClearTargetList()
	end
end


-- Yell-based detection (used for some vehicle fights, etc.)
function ABC:OnMonsterMessage(type, msg)

	-- Pull detection
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		for k, v in pairs(ABC.enc) do
			if v.pull and v.pull.type == "yell" and v.pull.yells then
				for kk, vv in pairs(v.pull.yells) do
					if vv == msg then
						self:Debug("Pull detected")
						ABC:StartFight(v, 0)
					end
				end
			end
		end
	end

	-- Defeat detection
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		for k, v in pairs(ABC.fight) do
			if v.defeat and v.defeat.type == "yell" and v.defeat.yells then
				for kk, vv in pairs(v.defeat.yells) do
					if vv == msg and v.inCombat then
						self:Debug("Defeat detected")
						ABC:EndFight(v)
					end
				end
			end
		end
	end

end
	


--[[
      WIPE DETECTION -----------------------------------------------------------------------------------------------------------
  ]]

function ABC:CheckForWipe()
	local wipe = true
	local uId = ((GetNumRaidMembers() == 0) and "party") or "raid"
	for i = 0, math.max(GetNumRaidMembers(), GetNumPartyMembers()) do
		local id = (i == 0 and "player") or uId..i
		if UnitAffectingCombat(id) and not UnitIsDeadOrGhost(id) then
			wipe = false
			break
		end
	end
	return wipe
end



--[[
      DEFEAT DETECTION ---------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnMobKill(id)
	for k, v in pairs(ABC.fight) do
		if v.defeat and v.defeat.type == "death" then
			if not v.dead then v.dead = {} end
			if v.defeat.npcs then
				for kk, vv in pairs(v.defeat.npcs) do
					if id == vv then
						v.dead[vv] = true
					end
				end
				local allDead = true
				for kk, vv in pairs(v.defeat.npcs) do
					if not v.dead[vv] then
						allDead = false
						break
					end
				end
				if allDead then
					if v.onDefeat and type(v.onDefeat) == "function" then v.onDefeat(v) end
					self:Debug("Defeat detected")
					self:EndFight(v)
				end
			end
		end
	end
end



--[[
      FIGHT LOGIC --------------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnTimer()
	self:UpdateRaid()
	for k, v in pairs(ABC.fight) do
		if v.onHeartbeat and type(v.onHeartbeat) == "function" then v.onHeartbeat(v) end
	end
end


function ABC:OnPower()
	for k, v in pairs(ABC.fight) do
		if v.onPower and type(v.onPower) == "function" then v.onPower(v) end
	end
end


function ABC:UpdateRaid()
	if UnitInRaid("player") then
	
		for k, v in pairs(ABC.raid) do
			ABC.raid[k].valid = nil
		end
	
		-- Gather data about the raid
		local numRaid = GetNumRaidMembers()
		for i = 1, numRaid do
			local playerName, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
			if playerName then
				if not ABC.raid[playerName] then ABC.raid[playerName] = {} end
				if not online and ABC.raid[playerName].online then
					for k, v in pairs(ABC.fight) do
						if v.onDisconnect and type(v.onDisconnect) == "function" then v.onDisconnect(v, playerName) end
					end
				end
				ABC.raid[playerName].rank = rank
				ABC.raid[playerName].subgroup = subgroup
				ABC.raid[playerName].level = level
				ABC.raid[playerName].class = class
				ABC.raid[playerName].zone = zone
				ABC.raid[playerName].online = online
				ABC.raid[playerName].isDead = isDead
				ABC.raid[playerName].role = role
				ABC.raid[playerName].isML = isML
				ABC.raid[playerName].valid = true
				ABC.raid[playerName].index = i
			end -- valid player at that index
		end -- player loop
		
		-- Delete players no longer in the raid
		for k, v in pairs(ABC.raid) do
			if not v.valid then
				ABC.raid[k] = nil
			end
		end
		
		-- Check for a wipe
		if ABC.isEnabled then
			for k, v in pairs(ABC.fight) do
				if self:CheckForWipe() and not v.possibleWipe then
					v.possibleWipe = true
					local time = 5
					if v.wipeTime then time = v.wipeTime end
					self:Debug(format("Possible wipe detected. Checking again in %d seconds.", time))
					self:ScheduleTimer(function()
						if v.possibleWipe then
							if ABC:CheckForWipe() then
								v.possibleWipe = false
								if v.onWipe and type(v.onWipe) == "function" then v.onWipe(v) end
								self:Debug("Wipe detected")
								ABC:EndFight(v)
							end
						end
					end, time)
				end -- wipe is probable
			end -- encounter loop
		end -- isEnabled
		
	else
		ABC.raid = {}
	end -- we're in a raid
end


--- Cancels a possible wipe. A "possible wipe" occurs when ABC detects that the entire raid is
-- either dead, or not affecting combat. Rather than wiping immediately, ABC flags a "possible wipe"
-- and waits a certain number of seconds before checking again. If the second check also indicates
-- a wipe, ABC actually executes the wipe.
--
-- CancelPossibleWipe can be called during the delay between the first and second checks. It will
-- cancel the second check. This is useful for encounters (such as The Lich King) that contain
-- a very long period where the entire raid is dead.
function ABC:CancelPossibleWipe(fight)
	fight.possibleWipe = false
end


--- Loads any load-on-demand add-ons that requested to be loaded for the current (or designated) zone.
-- @param zone (Optional) If specified, LoadZone loads any add-ons belonging to the specified zone.
-- This parameter is expected to be in the enUS locale. If you don't specify this, uses the current zone.
-- @param now (Optional) Causes load-on-demand add-ons to be loaded immediately. Without this parameter,
-- the add-ons would be loaded after a 3 second delay.
function ABC:LoadZone(zone, now)
	if zone then ABC.testZone = zone end
	for k, v in pairs(self.modules) do
		for kk, vv in pairs(v.zone) do
			if (GetRealZoneText() == zonenames[vv] or (zone and zone == zonenames[vv])) and v.addon and not IsAddOnLoaded(v.addon) then
				local inInstance, instanceType = IsInInstance()
				if (inInstance and instanceType == "raid") or zone then
					if now then
						LoadAddOn(v.addon)
					else
						self:ScheduleTimer(function() LoadAddOn(v.addon) end, 3)
					end
				end
			end
		end
	end
end


--- Starts the selected fight.
-- @param fight The fight definition table to start, or the fight's unique key string.
-- @param startDelay (Optional) The delay in seconds between when the fight should have
-- started, and when it actually started.
-- @param silent (Optional) Used for debug mode only. Instructs the function to not output any debug text.
function ABC:StartFight(fight, startDelay, silent)
	if type(fight) == "string" then
		if ABC.enc[fight] then
			fight = ABC.enc[fight]
		else
			error(format("Unable to locate fight: %s", fight))
			return
		end
	end
	
	if not fight then
		error("No fight specified")
		return
	end
	
	if self:IsFightStarted(fight) then return end
	if not startDelay then startDelay = 0 end

	if fight.enabled ~= false and ABC.isEnabled then
		-- Initialize fight
		fight.inCombat = true
		fight.startDelay = startDelay
		fight.dead = {}
		fight.possibleWipe = false
		
		-- Start heartbeat and register for events if this is the first fight running
		if not ABC.testing and self:GetNumActiveFights() <= 0 then
			self:CancelAllTimers()
			self:ScheduleRepeatingTimer("OnTimer", POLLING_FREQUENCY)
			self:RegisterEvent("UNIT_TARGET", "OnTarget")
			self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnCombatEvent")
			self:RegisterEvent("UNIT_POWER_BAR_SHOW", "OnPower")
			self:RegisterEvent("UNIT_POWER_BAR_HIDE", "OnPower")
			self:RegisterEvent("UNIT_POWER", "OnPower")
			self:RegisterEvent("UNIT_MAXPOWER", "OnPower")
		end
		
		-- Add the fight to our table of fights
		ABC.fight[fight._key] = fight

		if not silent then
			local heroic = ""
			if ABC:IsHeroicRaid() then heroic = " (heroic)" end
			self:Debug(format("Initiating fight: %s"..heroic, fight.name or "Name not specified in encounter definition"))
			self:Debug(format("%d fights running", self:GetNumActiveFights()))
		end
		
		if fight.onStart and type(fight.onStart) == "function" then fight.onStart(fight) end
		
		return true -- indicate success to callers
	end
end


--- Returns true if the selected fight is currently active, false if not.
-- @param fight The fight object. This should be the fight table, not the fight's key string.
function ABC:IsFightStarted(fight)
	if fight._key then
		for k, v in pairs(ABC.fight) do
			if v._key == fight._key then return true end
		end
	end
	return false
end


--- Returns the number of fights currently active.
function ABC:GetNumActiveFights()
	local i = 0
	for k, v in pairs(ABC.fight) do
		i = i + 1
	end
	return i
end


--- Immediately ends the selected fight.
-- @param fight The fight object. This should be the fight table, not the fight's key string.
-- @param silent (Optional) Used for debug mode only. Instructs the function to not output any debug text.
function ABC:EndFight(fight, silent)
	if not fight or not ABC:IsFightStarted(fight) then return end
	
	if ABC:GetNumActiveFights() <= 1 then
		self:CancelAllTimers()
		self:UnregisterEvent("UNIT_TARGET")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:UnregisterEvent("UNIT_POWER_BAR_SHOW")
		self:UnregisterEvent("UNIT_POWER_BAR_HIDE")
		self:UnregisterEvent("UNIT_POWER")
		self:UnregisterEvent("UNIT_MAXPOWER")
	end
	
	fight.inCombat = false
	if fight.onEnd and type(fight.onEnd) == "function" then fight.onEnd(ABC.fight) end
	ABC.fight[fight._key] = nil
	
	if not silent then
		self:Debug(format("Ending fight: %s", fight.name or "Name not specified in encounter definition"))
		self:Debug(format("%d fights running", self:GetNumActiveFights()))
	end
end


--- Immediately ends all fights.
-- @param silent (Optional) Used for debug mode only. Instructs the function to not output any debug text.
function ABC:EndAllFights(silent)
	for k, v in pairs(ABC.fight) do
		self:EndFight(v, silent)
	end
end



--[[
      API ----------------------------------------------------------------------------------------------------------------------
  ]]

--- If debug mode is on, prints something using ABC:Print().
-- Turn debug mode on by typing /allarabosscore debug.
function ABC:Debug(...)
	if AllaraBossCoreDB.debugMode then self:Print(...) end
end


--- Registers a table of encounters. The keys in the table are used as the unique fight keys.
-- Each table entry is an encounter definition table.
-- @param encounters A table of encounter definitions (which are themselves tables).
function ABC:RegisterEncounters(encounters)
	for k, v in pairs(encounters) do
		self:RegisterEncounter(k, v)
	end
end


--- Registers an encounter definition. Your encounter will not be validated for errors.
-- @param name The unique key string for this encounter. This key must be unique across
-- all add-ons that use ABC. If a duplicate key is registered twice, an error will be thrown.
-- @param definition The encounter definition table. See http://www.wowace.com/addons/allara-boss-core/pages/how-to-write-an-add-on-that-uses-abc/
-- for details about this format.
function ABC:RegisterEncounter(name, definition)
	if self:ValidateEncounter(name, definition) then
		for k, v in pairs(ABC.enc) do
			if k == name then
				error(format("An encounter definition with that key has already been registered. Keys must be unique across all add-ons that use ABC. (%s)", name))
				return
			end
		end
		definition._key = name
		ABC.enc[name] = definition
		if definition.zone then ABC.zones[zonenames[definition.zone]] = true end
		ABC.enc[name].inCombat = false
	end
end


function ABC:ValidateEncounter(name, definition)
	return true
end


--- Returns true if the current raid instance is in heroic mode.
function ABC:IsHeroicRaid()
	if ABC.forceHeroic ~= nil then return ABC.forceHeroic end
	local diff = GetInstanceDifficulty()
	local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
	if isDynamic then
		if dynamicDifficulty == 1 then return true else return false end
	end
	if diff == 3 or diff == 4 then return true else return false end
end


--- Sends arbitrary data to other players in your raid. When received,
-- your encounter definition will receive the onSync event. The data is
-- only sent to your encounter, not any other encounters that may be running.
-- @param fight The fight object. This should be the fight table, not the fight's key string.
-- If calling from within the encounter, simply pass "self" as this parameter.
-- @param object Any arbitrary object to send. This can be any value, or a table.
function ABC:Sync(fight, object)
	if fight and fight._key and object then
		self:SendCommMessage(commPrefix, self:Serialize({ c = "SYNC", addon = fight._key, data = object }), "RAID")
	end
end


--- Returns the NPC ID of a given unit ID
-- @param unit The unit ID (i.e. player, raid1target, etc.)
function ABC:GetUnitCreatureId(unit)
	if unit then
		local guid = UnitGUID(unit)
		if guid then
			return (guid and tonumber(guid:sub(7, 10), 16)) or 0
		end
	end
end


--- Returns the NPC ID of a given unit from it's GUID.
-- @param guid The unit's GUID.
function ABC:GetCIDFromGUID(guid)
	if guid then
		return (guid and tonumber(guid:sub(7, 10), 16)) or 0
	end
end
